<?php

/*
 * Author: Hugo Hallqvist <hugo@dokad.se>
 * Author: Kostyantyn Gushtin <kostya.gushtin@n-ix.com.ua>
 *
 * Daemon responsible for loading loadfiles into the database, and moving
 * them to local archives afterwards.
 *
 * */

// Make sure all errors are displayed
error_reporting(E_ALL);
ini_set("display_errors", "On");
declare(ticks = 1);

$db_dont_connect = true;
//require_once("database.php");
require_once("logger.inc.php");
require_once("daemon.inc.php");

define("NO_ARCHIVE", "NO_ARCHIVE");
define("DEFAULT_LOADFILE_ARCHIVE", "/var/log/oslogd/archive");
define("DEFAULT_ERRORFILE_ARCHIVE", "/var/log/oslogd/error");
define("DEFAULT_SPOOL_DIR", "/var/log/oslogd/spool");

// line mode defines
define("ERR_LINE", 0);
define("OK_LINE", 1);
class LoadDaemon extends Daemon
{
	private $timeout = 30;
	private $laststand = 0;

	function __construct()
	{
		parent::__construct("op5-log-loader", "op5-log-loader");
	}

	function daemonMain()
	{
		while (! $this->getShouldExit()) {
			$this->flushQueue(DEFAULT_SPOOL_DIR);

			sleep($this->timeout);
		}
	}

	function ensureDir($dirnam, $mode, $recursive)
	{
		// Try to ensure that the directory exists
		if (! is_dir($dirnam)) {
			if (! mkdir($dirnam, $mode, $recursive)) {
				return false;
			}
		}
		return true;
	}

	function dbLoad($filename)
	{
		$retval = true;
		$move_dir = NO_ARCHIVE;
		if (! $this->loadFile($filename)) {
			Logger::log("Error caught in load_file(%s), falling back to line mode", $filename);
			if (! $this->loadLines($filename)) {
				// line mode fallback failed, nothing else to do but move the entire minute to
				// error archive
				Logger::log("Can't process file %s. Moved it to error directory %s", $filename, DEFAULT_ERRORFILE_ARCHIVE);
				$move_dir = DEFAULT_ERRORFILE_ARCHIVE;
			}
			else {
				unlink($filename);
			}
		}
		else
			$move_dir = DEFAULT_LOADFILE_ARCHIVE;


		if (strcmp($move_dir, NO_ARCHIVE) != 0) {
			// Try to ensure that the directory exists
			if (! $this->ensureDir($move_dir, 0700, true)) {
				Logger::log("loadfile_archive (%s) did not exist and creation failed.", DEFAULT_LOADFILE_ARCHIVE);
				$retval = false;
				break;
			}

			// Move file into local archive storage
			$newfilename = sprintf("%s/%s", $move_dir, basename($filename));
			if (!rename($filename, $newfilename)) {
				Logger::log("rename(%s, %s) failed!", $filename, $newfilename);
				$retval = false;
				break;
			}
		}
		return $retval;

	}

	function loadFile($filename)
	{

		if (!pg_connect("dbname=op5lsu user=op5lsu"))
		{
			Logger::log("Please, start PostgreSQL.");
			return false;
		}

		exec('/opt/logserver/connector/nl-fixer ' . $filename);

		$log_time = explode("_", substr(basename($filename), 4));
		$query = sprintf("select load_to_table_function('%s-%s-%s', '%s:00:00', '$filename');", $log_time[0], $log_time[1], $log_time[2], $log_time[3]);
		unset($log_time);

		Logger::debug("Trying to execute pg_query(%s)", $query);

		$retval = true;
		if (!$res = pg_query($query)) {
			Logger::log("Got error from pgsql: %s", str_replace("", "", pg_last_error()));
			$retval = false;
		}

		$status = pg_fetch_result($res, 0);
		if ($status == "f")
		{
			$retval = false;
		}

		pg_close();

		return $retval;
	}

	function loadLines($filename)
	{
		if (!pg_connect("dbname=op5lsu user=op5lsu"))
		{
			Logger::log("Please, start PostgreSQL.");
			return false;
		}
		if(!$in_handle = fopen($filename, 'r'))
		{
			Logger::log("loadLines(%s) failed to open input log file for reading", $filename);
			return false;
		}

		$log_time = explode("_", substr(basename($filename), 4));


		foreach (array(DEFAULT_ERRORFILE_ARCHIVE, DEFAULT_LOADFILE_ARCHIVE) as $arch_dir) {
			if (! $this->ensureDir($arch_dir, 0700, true)) {
				Logger::log("loadfile_archive (%s) did not exist and creation failed.", $arch_dir);
				return false;
			}
		}

		$err_filename = sprintf("%s/%s", DEFAULT_ERRORFILE_ARCHIVE, basename($filename));
		if(!$err_handle = fopen($err_filename, 'a'))
		{
			Logger::log("Failed to open (%s) for writing, giving up on line mode", $err_filename);
			return false;
		}

		$ok_filename = sprintf("%s/%s", DEFAULT_LOADFILE_ARCHIVE, basename($filename));
		if(!$ok_handle = fopen($ok_filename, 'a'))
		{
			Logger::log("Failed to open (%s) for writing, giving up on line mode", $ok_filename);
			return false;
		}
		$max_lines = 1000;
		$ok_lines = array();
		$err_lines = array();

		while (($line = fgets($in_handle)) != FALSE) {
			$v_arr = explode("", $line);

			if (count($v_arr) != 10) {
				// Missing field, don't even bother...
				$err_lines[] = $line;
				continue;
			}

			$format = "select insert_to_table_function('%s-%s-%s', '%s:00:00', '%s', '%s', '%s', '%s',";
			$format .= "'%s', '%s', '%s', '%s', '%s', '%s')";

			$query = sprintf($format, $log_time[0], $log_time[1], $log_time[2], $log_time[3],
				pg_escape_string($v_arr[0]), pg_escape_string($v_arr[1]), pg_escape_string($v_arr[2]),
				pg_escape_string($v_arr[3]), pg_escape_string($v_arr[4]), pg_escape_string($v_arr[5]),
				pg_escape_string($v_arr[6]), pg_escape_string($v_arr[7]), pg_escape_string($v_arr[8]),
				pg_escape_string($v_arr[9]));

			Logger::debug("Trying to execute query (%s)", $query);

			if (!$res = pg_query($query)) {
				Logger::log("load_line: Got error from pgsql: %s", str_replace("", "", pg_last_error()));
				$err_lines[] = $line;
			}
			else {
				$status = pg_fetch_result($res, 0);
				if ($status == "f")
				{
					$err_lines[] = $line;
				}
				else
					$ok_lines[] = $line;
			}

			if(count($err_lines) >= $max_lines) {
				foreach ($err_lines as $line) {
					$this->archiveLine($err_handle, $line);
				}
				$err_lines = array();
			}

			if(count($ok_lines) >= $max_lines) {
				foreach ($ok_lines as $line) {
					$this->archiveLine($ok_handle, $line);
				}
				$ok_lines = array();
			}

		}
		foreach ($err_lines as $line) {
			$this->archiveLine($err_handle, $line);
		}

		foreach ($ok_lines as $line) {
			$this->archiveLine($ok_handle, $line);
		}

		fclose($in_handle);
		fclose($err_handle);
		fclose($ok_handle);
		return true;
	}

	function testArgValidity($file)
	{

		$ret = false;
		if (is_file($file) && !is_link($file))
			if (is_readable($file) && is_writable($file))
			{
				$filename = basename($file);
				if (ereg("([0-9]{4})_([0-9]{2})_([0-9]{2})_([0-9]{2})_([0-9]{2})", $filename))
				{
					$dt = explode("_", substr($filename, 4));
					$tm = strtotime($dt[0] . "-" . $dt[1] . "-" . $dt[2] . " " . $dt[3] . ":" . $dt[4] . ":00");

					if ($tm != -1 && $tm != "")
						$ret = true;
				}
			}

		return $ret;

	}


	function flushQueue($dir)
	{
		if (! $this->ensureDir($dir, 0755, true)) {
			Logger::log("loadfile_spooldir (%s) did not exist and creation failed.", $dir);
			return false;
		}

		$dirfiles = glob($dir . "/cfs.*");

		$logfiles = array();
		foreach ($dirfiles as $filename)
		{
			if ($this->testArgValidity($filename))
				$logfiles[] = $filename;
		}

		if(count($logfiles) == 1)
		{
			$this->laststand++;

			if( $this->laststand * $this->timeout < 100 )
				return true;

		}
		else
		{
			sort($logfiles);
			array_pop($logfiles);
		}

		$this->laststand = 0;
		$retval = true;
		foreach($logfiles as $filename)
		{
			// Load file into pgsql
			$retval = $this->dbLoad($filename);
		}

		return $retval;
	}

	function archiveLine($handle, $line)
	{
		fwrite($handle, $line);
	}
}

$daemon = new LoadDaemon();
$daemon->run();

exit; // Main program done, functions below

?>
